<?php

/**
 * VirtFusion Direct Provisioning Module
 *
 * @link https://docs.virtfusion.com/integrations/blesta VirtFusion
 */
class VirtfusionDirectProvisioning extends Module
{

    /**
     * Initializes the module
     */
    public function __construct()
    {
        // Load the language required by this module
        Language::loadLang('virtfusion_direct_provisioning', null, dirname(__FILE__) . DS . 'language' . DS);

        // Load components required by this module
        Loader::loadComponents($this, ['Input']);

        // Load module config
        $this->loadConfig(dirname(__FILE__) . DS . 'config.json');

    }

    private function getApi($api_token, $hostname)
    {
        Loader::load(dirname(__FILE__) . DS . 'apis' . DS . 'virtfusion_api.php');

        return new VirtfusionApi($api_token, $hostname);
    }

    /**
     * Performs any necessary bootstraping actions
     */
    public function install()
    {
    }

    /**
     * Performs migration of data from $current_version (the current installed version)
     * to the given file set version. Sets Input errors on failure, preventing
     * the module from being upgraded.
     *
     * @param string $current_version The current installed version of this module
     */
    public function upgrade($current_version)
    {
    }

    /**
     * Performs any necessary cleanup actions. Sets Input errors on failure
     * after the module has been uninstalled.
     *
     * @param int $module_id The ID of the module being uninstalled
     * @param bool $last_instance True if $module_id is the last instance
     *  across all companies for this module, false otherwise
     */
    public function uninstall($module_id, $last_instance)
    {
    }

    /**
     * Returns the rendered view of the manage module page.
     *
     * @param mixed $module A stdClass object representing the module and its rows
     * @param array $vars An array of post data submitted to or on the manager module
     *  page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the manager module page
     */
    public function manageModule($module, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('manage', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        $this->view->set('module', $module);

        return $this->view->fetch();
    }

    /**
     * Returns the rendered view of the add module row page.
     *
     * @param array $vars An array of post data submitted to or on the add module
     *  row page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the add module row page
     */
    public function manageAddRow(array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('add_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        if (!empty($vars)) {
            // Set unset checkboxes
            $checkbox_fields = [];

            foreach ($checkbox_fields as $checkbox_field) {
                if (!isset($vars[$checkbox_field])) {
                    $vars[$checkbox_field] = 'false';
                }
            }
        }

        $this->view->set('vars', (object)$vars);

        return $this->view->fetch();
    }

    /**
     * Returns the rendered view of the edit module row page.
     *
     * @param stdClass $module_row The stdClass representation of the existing module row
     * @param array $vars An array of post data submitted to or on the edit
     *  module row page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the edit module row page
     */
    public function manageEditRow($module_row, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('edit_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        if (empty($vars)) {
            $vars = $module_row->meta;
        } else {
            // Set unset checkboxes
            $checkbox_fields = [];

            foreach ($checkbox_fields as $checkbox_field) {
                if (!isset($vars[$checkbox_field])) {
                    $vars[$checkbox_field] = 'false';
                }
            }
        }

        $this->view->set('vars', (object)$vars);

        return $this->view->fetch();
    }

    /**
     * Adds the module row on the remote server. Sets Input errors on failure,
     * preventing the row from being added. Returns a set of data, which may be
     * a subset of $vars, that is stored for this module row.
     *
     * @param array $vars An array of module info to add
     * @return array A numerically indexed array of meta fields for the module row containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     */
    public function addModuleRow(array &$vars)
    {
        $meta_fields = ['name', 'hostname', 'api_token'];
        $encrypted_fields = ['api_token'];

        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        $this->Input->setRules($this->getRowRules($vars));

        // Validate module row
        if ($this->Input->validates($vars)) {
            $vars['hostname'] = strtolower($vars['hostname']);
            // Build the meta data for this row
            $meta = [];
            foreach ($vars as $key => $value) {
                if (in_array($key, $meta_fields)) {
                    $meta[] = [
                        'key' => $key,
                        'value' => $value,
                        'encrypted' => in_array($key, $encrypted_fields) ? 1 : 0
                    ];
                }
            }

            return $meta;
        }
    }

    /**
     * Edits the module row on the remote server. Sets Input errors on failure,
     * preventing the row from being updated. Returns a set of data, which may be
     * a subset of $vars, that is stored for this module row.
     *
     * @param stdClass $module_row The stdClass representation of the existing module row
     * @param array $vars An array of module info to update
     * @return array A numerically indexed array of meta fields for the module row containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     */
    public function editModuleRow($module_row, array &$vars)
    {
        $meta_fields = ['name', 'hostname', 'api_token'];
        $encrypted_fields = ['api_token'];

        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        $this->Input->setRules($this->getRowRules($vars));

        // Validate module row
        if ($this->Input->validates($vars)) {
            $vars['hostname'] = strtolower($vars['hostname']);
            // Build the meta data for this row
            $meta = [];
            foreach ($vars as $key => $value) {
                if (in_array($key, $meta_fields)) {
                    $meta[] = [
                        'key' => $key,
                        'value' => $value,
                        'encrypted' => in_array($key, $encrypted_fields) ? 1 : 0
                    ];
                }
            }

            return $meta;
        }
    }

    /**
     * Builds and returns the rules required to add/edit a module row (e.g. server).
     *
     * @param array $vars An array of key/value data pairs
     * @return array An array of Input rules suitable for Input::setRules()
     */
    private function getRowRules(&$vars)
    {
        $rules = [
            'name' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.name.valid', true)
                ]
            ],
            'hostname' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.hostname.valid', true)
                ]
            ],
            'api_token' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.api_token.valid', true)
                ]
            ]
        ];

        return $rules;
    }


    /**
     * Determines which module row should be attempted when a service is provisioned
     * for the given group based upon the order method set for that group.
     *
     * @return int The module row ID to attempt to add the service with
     * @see Module::getGroupOrderOptions()
     */
    public function selectModuleRow($module_group_id)
    {
        if (!isset($this->ModuleManager)) {
            Loader::loadModels($this, ['ModuleManager']);
        }

        $group = $this->ModuleManager->getGroup($module_group_id);

        if ($group) {
            switch ($group->add_order) {
                default:
                case 'first':
                    break;
                case 'roundrobin':
                    break;
            }
        }
        return 0;
    }

    /**
     * Validates input data when attempting to add a package, returns the meta
     * data to save when adding a package. Performs any action required to add
     * the package on the remote server. Sets Input errors on failure,
     * preventing the package from being added.
     *
     * @param array An array of key/value pairs used to add the package
     * @return array A numerically indexed array of meta fields to be stored for this package containing:
     *
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function addPackage(array $vars = null)
    {
        // Set rules to validate input data
        $this->Input->setRules($this->getPackageRules($vars));

        // Build meta data to return
        $meta = [];
        if ($this->Input->validates($vars)) {
            if (!isset($vars['meta'])) {
                return [];
            }

            // Return all package meta fields
            foreach ($vars['meta'] as $key => $value) {
                $meta[] = [
                    'key' => $key,
                    'value' => $value,
                    'encrypted' => 0
                ];
            }
        }

        return $meta;
    }

    /**
     * Deletes the package on the remote server. Sets Input errors on failure,
     * preventing the package from being deleted.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function deletePackage($package)
    {
    }

    /**
     * Builds and returns rules required to be validated when adding/editing a package.
     *
     * @param array $vars An array of key/value data pairs
     * @return array An array of Input rules suitable for Input::setRules()
     */
    private function getPackageRules(array $vars)
    {
        // Validate the package fields
        $rules = [
            'hypervisor_group_id' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.hypervisor_group_id.valid', true)
                ]
            ],
            'default_ipv4' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.default_ipv4.valid', true)
                ]
            ],
            'package_id' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.package_id.valid', true)
                ]
            ]
        ];

        return $rules;
    }

    /**
     * Returns all fields used when adding/editing a package, including any
     * javascript to execute when the page is rendered with these fields.
     *
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to
     *  render as well as any additional HTML markup to include
     */
    public function getPackageFields($vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Hypervisor Group ID field
        $hypervisor_group_id = $fields->label(Language::_('VirtfusionDirectProvisioning.package_fields.hypervisor_group_id', true), 'virtfusion_direct_provisioning_hypervisor_group_id');
        $hypervisor_group_id->attach(
            $fields->fieldText(
                'meta[hypervisor_group_id]',
                (isset($vars->meta['hypervisor_group_id']) ? $vars->meta['hypervisor_group_id'] : null),
                ['id' => 'virtfusion_direct_provisioning_hypervisor_group_id']
            )
        );
        $fields->setField($hypervisor_group_id);

        // Set the Default IPv4 field
        $default_ipv4 = $fields->label(Language::_('VirtfusionDirectProvisioning.package_fields.default_ipv4', true), 'virtfusion_direct_provisioning_default_ipv4');
        $default_ipv4->attach(
            $fields->fieldText(
                'meta[default_ipv4]',
                (isset($vars->meta['default_ipv4']) ? $vars->meta['default_ipv4'] : null),
                ['id' => 'virtfusion_direct_provisioning_default_ipv4']
            )
        );
        $fields->setField($default_ipv4);

        // Set the Package ID field
        $package_id = $fields->label(Language::_('VirtfusionDirectProvisioning.package_fields.package_id', true), 'virtfusion_direct_provisioning_package_id');
        $package_id->attach(
            $fields->fieldText(
                'meta[package_id]',
                (isset($vars->meta['package_id']) ? $vars->meta['package_id'] : null),
                ['id' => 'virtfusion_direct_provisioning_package_id']
            )
        );
        $fields->setField($package_id);

        return $fields;
    }

    /**
     * Adds the service to the remote server. Sets Input errors on failure,
     * preventing the service from being added.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param array $vars An array of user supplied info to satisfy the request
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being added (if the current service is an addon service
     *  service and parent service has already been provisioned)
     * @param string $status The status of the service being added. These include:
     *  - active
     *  - canceled
     *  - pending
     *  - suspended
     * @return array A numerically indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function addService(
        $package,
        array $vars = null,
        $parent_package = null,
        $parent_service = null,
        $status = 'pending'
    )
    {
        // $this->Input->setErrors(['api' => ['response' => print_r($client, true)]]);
        // return;
        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        // Load the API
        $row = $this->getModuleRow();

        $api = $this->getApi($row->meta->api_token, $row->meta->hostname);

        // Get the fields for the service
        //$params = $this->getFieldsFromInput($vars, $package);

        // Validate the service-specific fields
        $this->validateService($package, $vars);

        if ($this->Input->errors()) {
            return;
        }

        // Only provision the service if 'use_module' is true
        if ($vars['use_module'] == 'true') {
            /**
             *
             * We need to check if a user exists in VirtFusion based on the extrelid
             *
             */
            // $service_fields = $this->serviceFieldsToObject($service->fields);

            $api->loadCommand('virtfusion_client');

            try {

                $server_api = new VirtfusionClient($api);
                $request = $server_api->check($vars['client_id'], []);


                if (isset($request['info'])) {
                    switch ($request['info']['http_code']) {
                        case 200:

                            $data = json_decode($request['response']);
                            /**
                             *
                             * A user already exists
                             *
                             */
                            break;

                        case 404:

                            Loader::loadModels($this, ['Clients']);

                            $client = $this->Clients->get($vars['client_id'], false);

                            $request = $server_api->create([
                                "name" => $client->first_name . ' ' . $client->last_name,
                                "email" => $client->email,
                                'extRelationId' => $vars['client_id']
                            ]);

                            if (isset($request['info'])) {
                                if ($request['info']['http_code'] !== 201) {
                                    $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                                    return;
                                }
                            } else {
                                $this->Input->setErrors(['api' => ['response' => 'Failed to get a response from the API. The action was unsuccessful.']]);
                                return;
                            }

                            $data = json_decode($request['response']);

                            break;

                        default:
                            $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                            return;
                    };


                    /**
                     *
                     * Create server
                     *
                     */

                    $api->loadCommand('virtfusion_server');

                    $server_api = new VirtfusionServer($api);

                    $request = $server_api->create([
                        "packageId" => $package->meta->package_id,
                        "userId" => $data->data->id,
                        "hypervisorId" => $package->meta->hypervisor_group_id,
                        "ipv4" => $package->meta->default_ipv4,
                    ]);

                    if ($request['info']['http_code'] !== 201) {
                        $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                        return;
                    }

                    $data = json_decode($request['response']);

                    return [
                        [
                            'key' => 'server_id',
                            'value' => $data->data->id,
                            'encrypted' => 0
                        ],
                        [
                            'key' => 'label',
                            'value' => '',
                            'encrypted' => 0
                        ]
                    ];

                } else {
                    $this->Input->setErrors(['api' => ['response' => 'Failed to get a response from the API. The action was unsuccessful.']]);
                    return;
                }

                return;
            } catch (Exception $e) {
                $this->Input->setErrors(['api' => ['response' => print_r($e->getMessage(), true)]]);
                return;
            }
        }

        return;

        // Return service fields
    }

    /**
     * Edits the service on the remote server. Sets Input errors on failure,
     * preventing the service from being edited.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $vars An array of user supplied info to satisfy the request
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being edited (if the current service is an addon service)
     * @return array A numerically indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function editService($package, $service, array $vars = null, $parent_package = null, $parent_service = null)
    {

        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        $service_fields = $this->serviceFieldsToObject($service->fields);

        $this->validateService($package, $vars, true);

        if ($this->Input->errors()) {
            return;
        }

        // Only update the service if 'use_module' is true
        if ($vars['use_module'] == 'true') {
        }

        // Return all the service fields
        $fields = ['server_id', 'label'];
        $encrypted_fields = [];
        $return = [];
        foreach ($fields as $field) {
            if (isset($vars[$field]) || isset($service_fields[$field])) {
                $return[] = [
                    'key' => $field,
                    'value' => $vars[$field] ?? $service_fields[$field],
                    'encrypted' => (in_array($field, $encrypted_fields) ? 1 : 0)
                ];
            }
        }

        return $return;
    }

    /**
     * Suspends the service on the remote server. Sets Input errors on failure,
     * preventing the service from being suspended.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being suspended (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function suspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {

            $api = $this->getApi($row->meta->api_token, $row->meta->hostname);
            $service_fields = $this->serviceFieldsToObject($service->fields);

            $api->loadCommand('virtfusion_server');

            try {

                $server_api = new VirtfusionServer($api);
                $request = $server_api->suspend($service_fields->server_id, []);

                $success = false;

                if (isset($request['info'])) {
                    if ($request['info']['http_code'] === 204) {
                        $success = true;
                    } else {
                        $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                    }

                } else {
                    $this->Input->setErrors(['api' => ['response' => 'Failed to get a response from the API. The action was unsuccessful.']]);
                }

                $this->log($row->meta->hostname . '| suspend', serialize($request), "output", $success);

                if (!$success) {
                    return;
                }
                return true;
            } catch (Exception $e) {
                // Nothing to do
                return;
            }
        }

        return null;
    }

    /**
     * Unsuspends the service on the remote server. Sets Input errors on failure,
     * preventing the service from being unsuspended.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being unsuspended (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function unsuspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {

            $api = $this->getApi($row->meta->api_token, $row->meta->hostname);
            $service_fields = $this->serviceFieldsToObject($service->fields);

            $api->loadCommand('virtfusion_server');

            try {

                $server_api = new VirtfusionServer($api);
                $request = $server_api->unsuspend($service_fields->server_id, []);

                $success = false;

                if (isset($request['info'])) {
                    if ($request['info']['http_code'] === 204) {
                        $success = true;
                    } else {
                        $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                    }

                } else {
                    $this->Input->setErrors(['api' => ['response' => 'Failed to get a response from the API. The action was unsuccessful.']]);
                }

                $this->log($row->meta->hostname . '| unsuspend', serialize($request), "output", $success);

                if (!$success) {
                    return;
                }
                return true;
            } catch (Exception $e) {
                // Nothing to do
                return;
            }

        }

        return null;
    }

    /**
     * Cancels the service on the remote server. Sets Input errors on failure,
     * preventing the service from being canceled.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being canceled (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function cancelService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {
            $api = $this->getApi($row->meta->api_token, $row->meta->hostname);
            $service_fields = $this->serviceFieldsToObject($service->fields);

            $api->loadCommand('virtfusion_server');

            try {

                $server_api = new VirtfusionServer($api);
                $request = $server_api->cancel($service_fields->server_id, []);

                $success = false;

                if (isset($request['info'])) {
                    if ($request['info']['http_code'] === 204) {
                        $success = true;
                    } else {
                        $this->Input->setErrors(['api' => ['response' => 'Received  a ' . $request['info']['http_code'] . ' http code from the API. The action was unsuccessful.']]);
                    }

                } else {
                    $this->Input->setErrors(['api' => ['response' => 'Failed to get a response from the API. The action was unsuccessful.']]);
                }

                $this->log($row->meta->hostname . '| cancel', serialize($request), "output", $success);

                if (!$success) {
                    return;
                }
                return true;
            } catch (Exception $e) {
                // Nothing to do
                return;
            }
        }
        return null;
    }

    /**
     * Attempts to validate service info. This is the top-level error checking method. Sets Input errors on failure.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param array $vars An array of user supplied info to satisfy the request
     * @return bool True if the service validates, false otherwise. Sets Input errors when false.
     */
    public function validateService($package, array $vars = null)
    {
        $this->Input->setRules($this->getServiceRules($vars));
        return $this->Input->validates($vars);
    }

    /**
     * Attempts to validate an existing service against a set of service info updates. Sets Input errors on failure.
     *
     * @param stdClass $service A stdClass object representing the service to validate for editing
     * @param array $vars An array of user-supplied info to satisfy the request
     * @return bool True if the service update validates or false otherwise. Sets Input errors when false.
     */
    public function validateServiceEdit($service, array $vars = null)
    {
        $this->Input->setRules($this->getServiceRules($vars, true));
        return $this->Input->validates($vars);
    }

    /**
     * Returns the rule set for adding/editing a service
     *
     * @param array $vars A list of input vars
     * @param bool $edit True to get the edit rules, false for the add rules
     * @return array Service rules
     */
    private function getServiceRules(array $vars = null, $edit = false)
    {
        // Validate the service fields
        $rules = [
            'server_id' => [
                'valid' => [
                    'if_set' => $edit,
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.server_id.valid', true)
                ]
            ],
            'label' => [
                'valid' => [
                    'if_set' => $edit,
                    'rule' => true,
                    'message' => Language::_('VirtfusionDirectProvisioning.!error.label.valid', true)
                ]
            ]
        ];

        // Unset irrelevant rules when editing a service
        if ($edit) {
            $edit_fields = [];

            foreach ($rules as $field => $rule) {
                if (!in_array($field, $edit_fields)) {
                    unset($rules[$field]);
                }
            }
        }

        return $rules;
    }

    /**
     * Updates the package for the service on the remote server. Sets Input
     * errors on failure, preventing the service's package from being changed.
     *
     * @param stdClass $package_from A stdClass object representing the current package
     * @param stdClass $package_to A stdClass object representing the new package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being changed (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function changeServicePackage(
        $package_from,
        $package_to,
        $service,
        $parent_package = null,
        $parent_service = null
    )
    {
        if (($row = $this->getModuleRow())) {
        }

        return null;
    }


    /**
     * Fetches the HTML content to display when viewing the service info in the
     * admin interface.
     *
     * @param stdClass $service A stdClass object representing the service
     * @param stdClass $package A stdClass object representing the service's package
     * @return string HTML content containing information to display when viewing the service info
     */
    public function getAdminServiceInfo($service, $package)
    {
        $row = $this->getModuleRow();

        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('admin_service_info', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html']);

        $this->view->set('module_row', $row);
        $this->view->set('package', $package);
        $this->view->set('service', $service);
        $this->view->set('service_fields', $this->serviceFieldsToObject($service->fields));

        return $this->view->fetch();
    }

    /**
     * Returns all tabs to display to a client when managing a service whose
     * package uses this module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @return array An array of tabs in the format of method => title.
     *  Example: array('methodName' => "Title", 'methodName2' => "Title2")
     */
    public function getClientTabs($package)
    {
        return [
            'tabManage' => Language::_('VirtfusionDirectProvisioning.tabManage', true)
        ];
    }

    public function getAdminTabs($package)
    {
        return [
            'tabAdminManage' => Language::_('VirtfusionDirectProvisioning.tabManage', true)
        ];
    }

    /**
     * tabManage
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabManage(
        $package,
        $service,
        array $get = null,
        array $post = null,
        array $files = null
    )
    {
        $this->view = new View('tabManage', 'default');
        $this->view->base_uri = $this->base_uri;

        Loader::loadHelpers($this, ['Form', 'Html']);

        $service_fields = $this->serviceFieldsToObject($service->fields);

        if ($_POST) {
            if (property_exists($service_fields, 'server_id')) {
                if (is_numeric($service_fields->server_id)) {
                    if (($row = $this->getModuleRow())) {
                        $api = $this->getApi($row->meta->api_token, $row->meta->hostname);

                        $api->loadCommand('virtfusion_server');

                        $server_api = new VirtfusionServer($api);
                        $request = $server_api->fetchToken($service_fields->server_id, $service->client_id, []);

                        if (isset($request['info'])) {
                            if ($request['info']['http_code'] === 200) {
                                $data = json_decode($request['response']);

                                header("Location: https://" . $row->meta->hostname . $data->data->authentication->endpoint_complete);
                                die();

                            }
                        }
                    }
                }
            }
            $this->Input->setErrors([['We couldn\'t log you in. Something went wrong.']]);
        }

        $this->view->set('service_fields', $service_fields);
        $this->view->set('service_id', $service->id);
        $this->view->set('client_id', $service->client_id);
        $this->view->set('vars', (isset($vars) ? $vars : new stdClass()));

        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);
        return $this->view->fetch();
    }

    public function tabAdminManage(
        $package,
        $service,
        array $get = null,
        array $post = null,
        array $files = null
    )
    {
        $this->view = new View('tabAdminManage', 'default');

        $this->view->base_uri = $this->base_uri;
        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html']);

        $service_fields = $this->serviceFieldsToObject($service->fields);

        if ($_POST) {
            if (property_exists($service_fields, 'server_id')) {
                if (is_numeric($service_fields->server_id)) {
                    if (($row = $this->getModuleRow())) {
                        $api = $this->getApi($row->meta->api_token, $row->meta->hostname);

                        $api->loadCommand('virtfusion_server');

                        $server_api = new VirtfusionServer($api);
                        $request = $server_api->fetchToken($service_fields->server_id, $service->client_id, []);

                        if (isset($request['info'])) {
                            if ($request['info']['http_code'] === 200) {
                                $data = json_decode($request['response']);

                                header("Location: https://" . $row->meta->hostname . $data->data->authentication->endpoint_complete);
                                die();
                            }
                        }
                    }
                }
            }
            $this->Input->setErrors([['We couldn\'t log you in. Something went wrong.']]);
        }

        $this->view->set('service_fields', $service_fields);
        $this->view->set('service_id', $service->id);
        $this->view->set('client_id', $service->client_id);
        $this->view->set('vars', (isset($vars) ? $vars : new stdClass()));

        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);
        return $this->view->fetch();
    }

    /**
     * Fetches the HTML content to display when viewing the service info in the
     * client interface.
     *
     * @param stdClass $service A stdClass object representing the service
     * @param stdClass $package A stdClass object representing the service's package
     * @return string HTML content containing information to display when viewing the service info
     */
    public function getClientServiceInfo($service, $package)
    {
        $row = $this->getModuleRow();

        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('client_service_info', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'virtfusion_direct_provisioning' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html']);

        $this->view->set('module_row', $row);
        $this->view->set('package', $package);
        $this->view->set('service', $service);
        $this->view->set('service_fields', $this->serviceFieldsToObject($service->fields));

        return $this->view->fetch();
    }

    /**
     * Returns all fields to display to an admin attempting to add a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getAdminAddFields($package, $vars = null)
    {

        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();
        /*
              // Set the Server ID field
              $server_id = $fields->label(Language::_('VirtfusionDirectProvisioning.service_fields.server_id', true), 'virtfusion_direct_provisioning_server_id');
              $server_id->attach(
                  $fields->fieldText(
                      'server_id',
                      (isset($vars->server_id) ? $vars->server_id : null),
                      ['id' => 'virtfusion_direct_provisioning_server_id']
                  )
              );
              $fields->setField($server_id);
      */
        // Set the Label field
        $label = $fields->label(Language::_('VirtfusionDirectProvisioning.service_fields.label', true), 'virtfusion_direct_provisioning_label');
        $label->attach(
            $fields->fieldText(
                'label',
                (isset($vars->label) ? $vars->label : null),
                ['id' => 'virtfusion_direct_provisioning_label']
            )
        );
        $fields->setField($label);
        return $fields;
    }

    /**
     * Returns all fields to display to an admin attempting to edit a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getAdminEditFields($package, $vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Server ID field
        $server_id = $fields->label(Language::_('VirtfusionDirectProvisioning.service_fields.server_id', true), 'virtfusion_direct_provisioning_server_id');
        $server_id->attach(
            $fields->fieldText(
                'server_id',
                (isset($vars->server_id) ? $vars->server_id : null),
                ['id' => 'virtfusion_direct_provisioning_server_id']
            )
        );
        $fields->setField($server_id);

        // Set the Label field
        $label = $fields->label(Language::_('VirtfusionDirectProvisioning.service_fields.label', true), 'virtfusion_direct_provisioning_label');
        $label->attach(
            $fields->fieldText(
                'label',
                (isset($vars->label) ? $vars->label : null),
                ['id' => 'virtfusion_direct_provisioning_label']
            )
        );
        $fields->setField($label);

        return $fields;
    }
}
